"""
Convenience interface for NSArray/NSMutableArray
"""
__all__ = ()

import collections.abc
import sys

from objc._convenience import addConvenienceForClass, container_unwrap, container_wrap
from objc._objc import _C_ID, _C_NSInteger
from objc._objc import _NSNotFound as NSNotFound
from objc._objc import lookUpClass, registerMetaDataForSelector

NSArray = lookUpClass("NSArray")
NSMutableArray = lookUpClass("NSMutableArray")

collections.abc.Sequence.register(NSArray)
collections.abc.MutableSequence.register(NSMutableArray)


registerMetaDataForSelector(
    b"NSObject",
    b"sortUsingFunction:context:",
    {
        "arguments": {
            2: {
                "callable": {
                    "retval": {"type": _C_NSInteger},
                    "arguments": {
                        0: {"type": _C_ID},
                        1: {"type": _C_ID},
                        2: {"type": _C_ID},
                    },
                },
                "callable_retained": False,
            },
            3: {"type": _C_ID},
        }
    },
)


def _ensure_array(anArray):
    """Return *anArray* as a list, tuple or NSArray"""
    if not isinstance(anArray, (NSArray, list, tuple)):
        anArray = list(anArray)
    return anArray


def nsarray_reverse(self):
    """Reverse an array"""
    begin = 0
    end = len(self) - 1
    while begin < end:
        self.exchangeObjectAtIndex_withObjectAtIndex_(begin, end)
        begin += 1
        end -= 1


def nsarray_extend(self, anArray):
    for item in anArray:
        self.addObject_(container_wrap(item))


_index_sentinel = object()


def nsarray_index(self, item, start=0, stop=_index_sentinel):
    if start == 0 and stop is _index_sentinel:
        res = self.indexOfObject_(container_wrap(item))
        if res == NSNotFound:
            raise ValueError(f"{type(self).__name__}.index(x): x not in list")

    else:
        itemcount = self.count()
        if start < 0:
            start = itemcount + start
            if start < 0:
                start = 0

        if stop is not _index_sentinel:
            if stop < 0:
                stop = itemcount + stop
                if stop < 0:
                    stop = 0
        else:
            stop = itemcount

        if itemcount == 0:
            raise ValueError(f"{type(self).__name__}.index(x): x not in list")

        if start >= itemcount:
            raise ValueError(f"{type(self).__name__}.index(x): x not in list")

        if stop >= itemcount:
            stop = itemcount - 1

        if stop <= start:
            ln = 0

        else:
            ln = stop - start

        if ln == 0:
            raise ValueError(f"{type(self).__name__}.index(x): x not in list")

        if ln > sys.maxsize:  # pragma: no cover
            ln = sys.maxsize

        res = self.indexOfObject_inRange_(item, (start, ln))
        if res == NSNotFound:
            raise ValueError(f"{type(self).__name__}.index(x): x not in list")

    return res


def nsarray_insert(self, idx, item):
    if idx < 0:
        idx += self.count()
        if idx < 0:
            idx = 0
    if idx >= self.count():
        self.addObject_(item)
        return
    self.insertObject_atIndex_(container_wrap(item), idx)


def nsarray__getitem__(self, idx):
    if isinstance(idx, slice):
        start, stop, step = idx.indices(len(self))
        return [self[i] for i in range(start, stop, step)]

    elif not isinstance(idx, int):
        raise TypeError("index must be a number")

    if idx < 0:
        idx += len(self)
        if idx < 0:
            raise IndexError("list index out of range")

    return container_unwrap(self.objectAtIndex_(idx), RuntimeError)


def nsarray__delitem__(self, idx):
    if isinstance(idx, slice):
        start, stop, step = idx.indices(self.count())
        if step == 1:
            if start > stop:
                # Nothing to remove
                return

            return self.removeObjectsInRange_((start, stop - start))

        r = reversed(range(start, stop, step))
        for i in r:
            self.removeObjectAtIndex_(i)
        return

    if idx < 0:
        idx += self.count()
        if idx < 0:
            raise IndexError("list index out of range")

    self.removeObjectAtIndex_(idx)


def nsarray_pop(self, idx=-1):
    length = self.count()
    if length <= 0:
        raise IndexError("pop from empty list")

    elif idx >= length or (idx + length) < 0:
        raise IndexError("pop index out of range")

    elif idx < 0:
        idx += len(self)

    rval = self.objectAtIndex_(idx)
    self.removeObjectAtIndex_(idx)
    return rval


def nsarray_remove(self, obj):
    idx = self.indexOfObject_(obj)
    if idx == NSNotFound:
        raise ValueError(f"{type(self).__name__}.remove(x): x not in list")
    self.removeObjectAtIndex_(idx)


index_error_message = "list indices must be integers or slices"


def nsarray__setitem__(self, idx, anObject):
    if isinstance(idx, slice):
        start, stop, step = idx.indices(self.count())
        if step >= 0:
            if stop <= start:
                # Empty slice: insert values
                stop = start

        anObject = _ensure_array(anObject)

        if step == 1:
            return self.replaceObjectsInRange_withObjectsFromArray_(
                (start, stop - start), anObject
            )

        slice_len = len(range(start, stop, step))
        if slice_len != len(anObject):
            raise ValueError(
                "Replacing extended slice with %d elements by %d elements"
                % (slice_len, len(anObject))
            )

        if step > 0:
            # NOTE: 'anObject' cannot be 'self' because assigning to an extended
            # slice cannot change the size of 'self' and slep 1 is handled earlier.
            toAssign = anObject

            for inIdx, outIdx in enumerate(range(start, stop, step)):
                self.replaceObjectAtIndex_withObject_(outIdx, toAssign[inIdx])

        # slice.indexes already catches this:
        # elif step == 0:
        #    raise ValueError("Step 0")

        else:
            if anObject is self:
                toAssign = list(anObject)
            else:
                toAssign = anObject

            for inIdx, outIdx in enumerate(range(start, stop, step)):
                self.replaceObjectAtIndex_withObject_(outIdx, toAssign[inIdx])

    elif not isinstance(idx, int):
        raise TypeError(index_error_message)

    else:
        if idx < 0:
            idx += self.count()
            if idx < 0:
                raise IndexError("list index out of range")

        self.replaceObjectAtIndex_withObject_(idx, anObject)


def nsarray_add(self, other):
    result = NSMutableArray(self)
    result.addObjectsFromArray_(_ensure_array(other))
    return result


def nsarray_radd(self, other):
    result = NSMutableArray(other)
    result.addObjectsFromArray_(self)
    return result


def nsarray_mul(self, other):
    """
    This tries to implement anNSArray * N
    somewhat efficiently (and definitely more
    efficient that repeated appending).
    """
    result = NSMutableArray.array()

    if other <= 0:
        return result

    n = 1
    tmp = self
    while other:
        if (other & n) != 0:
            result.addObjectsFromArray_(tmp)
            other -= n

        if other:
            n <<= 1
            tmp = tmp.arrayByAddingObjectsFromArray_(tmp)

    return result


def nsarray_new(cls, sequence=None):
    if not sequence:
        return NSArray.array()

    elif isinstance(sequence, str):
        return NSArray.arrayWithArray_(list(sequence))

    else:
        if not isinstance(sequence, (list, tuple)):
            return NSArray.arrayWithArray_(list(sequence))

        return NSArray.arrayWithArray_(sequence)


def nsmutablearray_new(cls, sequence=None):
    if not sequence:
        return NSMutableArray.array()

    elif isinstance(sequence, str):
        return NSMutableArray.arrayWithArray_(list(sequence))

    else:
        if type(sequence) not in (list, tuple):
            return NSMutableArray.arrayWithArray_(list(sequence))

        # This is only valid when ``sequence`` is an built-in list or tuple,
        # otherwise arrayWithArray might access the sequence differently
        # then expected from a Python sequence initializer.
        return NSMutableArray.arrayWithArray_(sequence)


def nsarray__contains__(self, elem):
    return bool(self.containsObject_(container_wrap(elem)))


def nsarray_append(self, anObject):
    self.addObject_(container_wrap(anObject))


def nsarray_clear(self):
    self.removeAllObjects()


def nsarray_sort(self, key=lambda x: x, reverse=False):
    if reverse:

        def sort_func(a, b, _):
            a = key(a)
            b = key(b)
            if a < b:
                return 1

            elif b < a:
                return -1

            else:
                return 0

    else:

        def sort_func(a, b, _):
            a = key(a)
            b = key(b)

            if a < b:
                return -1
            elif b < a:
                return 1
            else:
                return 0

    self.sortUsingFunction_context_(sort_func, None)


def nsarray__len__(self):
    return self.count()


# NOTE: 'no cover' because call of the system array
# classes are subclasses of NSMutableArray.
def nsarray__copy__(self):  # pragma: no cover
    return self.copy()


def nsarray__iter__(self):
    return iter(self.objectEnumerator())


addConvenienceForClass(
    "NSArray",
    (
        ("__new__", staticmethod(nsarray_new)),
        ("__add__", nsarray_add),
        ("__radd__", nsarray_radd),
        ("__mul__", nsarray_mul),
        ("__rmul__", nsarray_mul),
        ("__len__", nsarray__len__),
        ("__contains__", nsarray__contains__),
        ("__getitem__", nsarray__getitem__),
        ("__copy__", nsarray__copy__),
        ("__iter__", nsarray__iter__),
        ("index", nsarray_index),
        ("remove", nsarray_remove),
        ("pop", nsarray_pop),
    ),
)

# XXX: See #334: Some part of Cocoa can load a category on subclasses of
# NSMutableArray that defines a pop method, which then interferes with
# nsarray_pop.
# The code below registers our own pop on all known subclasses of NSArray...
for cls in (
    "__NSArrayI",
    "__NSArrayM",
    "__NSArray0",
):
    addConvenienceForClass(
        cls,
        (("pop", nsarray_pop),),
    )


def nsmutablearray__copy__(self):
    return self.mutableCopy()


addConvenienceForClass(
    "NSMutableArray",
    (
        ("__new__", staticmethod(nsmutablearray_new)),
        ("__copy__", nsmutablearray__copy__),
        ("__setitem__", nsarray__setitem__),
        ("__delitem__", nsarray__delitem__),
        ("extend", nsarray_extend),
        ("append", nsarray_append),
        ("sort", nsarray_sort),
        ("insert", nsarray_insert),
        ("reverse", nsarray_reverse),
        ("clear", nsarray_clear),
    ),
)



from objc import _objc

__all__ = ["protocolNamed", "ProtocolError"]


class ProtocolError(_objc.error):
    __module__ = "objc"


PROTOCOL_CACHE = {}


def protocolNamed(name):
    """
    Returns a Protocol object for the named protocol. This is the
    equivalent of @protocol(name) in Objective-C.
    Raises objc.ProtocolError when the protocol does not exist.
    """
    try:
        return PROTOCOL_CACHE[name]
    except KeyError:
        pass
    for p in _objc.protocolsForProcess():
        pname = p.__name__
        PROTOCOL_CACHE.setdefault(pname, p)
        if pname == name:
            return p
    for cls in _objc.getClassList():
        for p in _objc.protocolsForClass(cls):
            pname = p.__name__
            PROTOCOL_CACHE.setdefault(pname, p)
            if pname == name:
                return p
    raise ProtocolError(f"protocol {name!r} does not exist", name)



"""
Generic conveniences for NSObject

The type is defined in Foundation, but NSObject is important
to the behavior of the bridge and therefore these conveniences
are kept in the core.
"""

__all__ = ()
import sys

from objc._convenience import addConvenienceForClass


def nsobject_hash(self, _max=sys.maxsize, _const=((sys.maxsize + 1) * 2)):
    rval = self.hash()
    if rval > _max:
        rval -= _const

    # -1 is not a valid hash in Python and hash(x) will
    # translate a hash of -1 to -2, so we might as well
    # do it here so that it's not too surprising..
    if rval == -1:
        rval = -2
    return int(rval)


def nsobject__eq__(self, other):
    if self.__is_magic():
        return self is other
    return bool(self.isEqualTo_(other))


def nsobject__ne__(self, other):
    if self.__is_magic():
        return self is not other
    return bool(self.isNotEqualTo_(other))


def nsobject__gt__(self, other):
    if self.__is_magic():
        return NotImplemented
    return bool(self.isGreaterThan_(other))


def nsobject__ge__(self, other):
    if self.__is_magic():
        return NotImplemented
    return bool(self.isGreaterThanOrEqualTo_(other))


def nsobject__lt__(self, other):
    if self.__is_magic():
        return NotImplemented
    return bool(self.isLessThan_(other))


def nsobject__le__(self, other):
    if self.__is_magic():
        return NotImplemented
    return bool(self.isLessThanOrEqualTo_(other))


class kvc:
    """
    Key-Value-Coding accessor for Cocoa objects.

    Both attribute access and dict-like indexing will attempt to
    access the requested item through Key-Value-Coding.
    """

    __slots__ = ("__object",)

    def __init__(self, value):
        self.__object = value

    def __repr__(self):
        return f"<KVC accessor for {self.__object!r}>"

    def __getattr__(self, key):
        try:
            return self.__object.valueForKey_(key)
        except KeyError as msg:
            if (
                hasattr(msg, "_pyobjc_info_")
                and msg._pyobjc_info_["name"] == "NSUnknownKeyException"
            ):
                raise AttributeError(key)

            raise

    def __setattr__(self, key, value):
        if not key.startswith("_"):
            return self.__object.setValue_forKey_(value, key)
        else:
            super().__setattr__(key, value)

    def __getitem__(self, key):
        if not isinstance(key, str):
            raise TypeError("Key must be string")

        return self.__object.valueForKey_(key)

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError("Key must be string")

        return self.__object.setValue_forKey_(value, key)


addConvenienceForClass(
    "NSObject",
    (
        ("__hash__", nsobject_hash),
        ("__eq__", nsobject__eq__),
        ("__ne__", nsobject__ne__),
        ("__gt__", nsobject__gt__),
        ("__ge__", nsobject__ge__),
        ("__lt__", nsobject__lt__),
        ("__le__", nsobject__le__),
        ("_", property(kvc)),
    ),
)



"""
Python representation for SIMD types in (Objective)C

This module only defines the vector types that are
used in framework bindings.
"""

import objc
import operator

__all__ = (
    "simd_int2",
    "simd_int3",
    "simd_int4",
    "simd_uint2",
    "simd_uint3",
    "simd_double2",
    "simd_double3",
    "simd_double4",
    "simd_float2",
    "simd_float3",
    "simd_float4",
    "simd_short2",
    "simd_ushort2",
    "simd_ushort3",
    "simd_ushort4",
    "simd_uchar16",
    "vector_float2",
    "vector_float3",
    "vector_float4",
    "vector_double2",
    "vector_double3",
    "vector_double4",
    "vector_short2",
    "vector_ushort2",
    "vector_ushort3",
    "vector_ushort4",
    "vector_int2",
    "vector_int3",
    "vector_int4",
    "vector_uint2",
    "vector_uint3",
    "vector_uchar16",
    "matrix_float2x2",
    "matrix_float3x3",
    "matrix_float4x3",
    "matrix_float4x4",
    "matrix_double4x4",
    "simd_quatf",
    "simd_quatd",
    "simd_float4x4",
)


def make_type(
    name,
    zero,
    cast_type,
    count,
    len2_type=None,
    len3_type=None,
    is_signed=True,
    limitrange=lambda x: x,
    typestr=None,
):
    assert count > 0
    if count in (3, 4):
        assert len2_type is not None
    if count == 4:
        assert len3_type is not None

    if count > 4:

        def __init__(self, *values):
            if len(values) == 0:
                self._values = [zero] * count

            elif len(values) == 1:
                v = values[0]
                v = cast_type(v)

                self._values = [v] * count
            elif len(values) != count:
                raise ValueError(values)

            else:
                self._values = [limitrange(cast_type(v)) for v in values]

    else:

        def __init__(self, *values):
            if len(values) == 0:
                self._values = [zero] * count
                return

            parts = []
            for p in values:
                if isinstance(p, self.__class__):
                    parts.extend(p._values)
                elif len2_type is not None and isinstance(p, len2_type):
                    parts.extend(p._values)
                elif len3_type is not None and isinstance(p, len3_type):
                    parts.extend(p._values)
                else:
                    parts.append(cast_type(p))

            if len(parts) == 1:
                self._values = parts * count

            elif len(parts) == count:
                self._values = parts

            else:
                raise ValueError(f"Expecting {count} values, got {len(parts)}")

            self._values = [limitrange(v) for v in self._values]

    def as_tuple(self):
        return tuple(self._values)

    def __repr__(self):
        return f"objc.simd.{name}({', '.join(map(str, self._values))})"

    def _objc_literal(self):
        return f"({name}){{{', '.join(map(str, self._values))}}}"

    def __len__(self):
        return count

    def __getitem__(self, idx):
        return self._values[idx]

    def __setitem__(self, idx, value):
        if not isinstance(idx, int):
            raise TypeError(idx)
        self._values[idx] = cast_type(value)

    def _cast_self(self, other):
        if not isinstance(other, self.__class__):
            try:
                return self.__class__(cast_type(other))
            except (ValueError, TypeError):
                return NotImplemented

        return other

    def __add__(self, other):
        other = self._cast_self(other)
        if other is NotImplemented:
            return NotImplemented

        return self.__class__(*(limitrange(x + y) for x, y in zip(self, other)))

    def __radd__(self, other):
        other = self._cast_self(other)
        if other is NotImplemented:
            return NotImplemented

        return self.__class__(*(limitrange(y + x) for x, y in zip(self, other)))

    def __mul__(self, other):
        other = self._cast_self(other)
        if other is NotImplemented:
            return NotImplemented

        return self.__class__(*(limitrange(x * y) for x, y in zip(self, other)))

    def __rmul__(self, other):
        other = self._cast_self(other)
        if other is NotImplemented:
            return NotImplemented

        return self.__class__(*(limitrange(y * x) for x, y in zip(self, other)))

    def __matmul__(self, other: "vector_float2") -> "vector_float2":
        if not isinstance(other, self.__class__):
            return NotImplemented

        return sum(x * y for x, y in zip(self, other))

    def __abs__(self):
        return self.__class__(*(abs(x) for x in self._values))

    def __eq__(self, other):
        other = self._cast_self(other)
        if other is NotImplemented:
            return NotImplemented

        return self._values == other._values

    def __ne__(self, other):
        other = self._cast_self(other)
        if other is NotImplemented:
            return NotImplemented

        return self._values != other._values

    def __lt__(self, other):
        other = self._cast_self(other)
        if other is NotImplemented:
            return NotImplemented

        return self._values < other._values

    def __le__(self, other):
        other = self._cast_self(other)
        if other is NotImplemented:
            return NotImplemented

        return self._values <= other._values

    def __gt__(self, other):
        other = self._cast_self(other)
        if other is NotImplemented:
            return NotImplemented

        return self._values > other._values

    def __ge__(self, other):
        other = self._cast_self(other)
        if other is NotImplemented:
            return NotImplemented

        return self._values >= other._values

    def __neg__(self):
        return self.__class__(*(-n for n in self._values))

    def __pos__(self):
        return self.__class__(self)

    class_dict = {
        "__slots__": ("_values",),
        "__init__": __init__,
        "as_tuple": as_tuple,
        "_cast_self": _cast_self,
        "__repr__": __repr__,
        "_objc_literal": _objc_literal,
        "__len__": __len__,
        "__getitem__": __getitem__,
        "__setitem__": __setitem__,
        "__add__": __add__,
        "__radd__": __radd__,
        "__mul__": __mul__,
        "__rmul__": __rmul__,
        "__matmul__": __matmul__,
        "__abs__": __abs__,
        "__eq__": __eq__,
        "__ne__": __ne__,
        "__lt__": __lt__,
        "__le__": __le__,
        "__gt__": __gt__,
        "__ge__": __ge__,
    }

    if is_signed:
        class_dict.update(
            {
                "__neg__": __neg__,
                "__pos__": __pos__,
            }
        )

    if 2 <= count <= 4:

        @property
        def x(self):
            return self._values[0]

        @x.setter
        def x(self, value):
            self._values[0] = cast_type(value)

        @property
        def y(self):
            return self._values[1]

        @y.setter
        def y(self, value):
            self._values[1] = cast_type(value)

        class_dict["x"] = x
        class_dict["y"] = y

    if count == 2:

        @property
        def xy(self):
            return self.__class__(*self._values)

        @xy.setter
        def xy(self, value):
            if not isinstance(value, self.__class__):
                raise TypeError(value)
            self._values[:] = value._values[:]

        class_dict["xy"] = xy

    elif count in (3, 4):

        @property
        def z(self):
            return self._values[2]

        @z.setter
        def z(self, value):
            self._values[2] = cast_type(value)

        class_dict["z"] = z

        @property
        def xy(self):
            return len2_type(*self._values[0:2])

        @xy.setter
        def xy(self, value):
            if not isinstance(value, len2_type):
                raise TypeError(value)
            self._values[0:2] = value._values[0:2]

        class_dict["xy"] = xy

        @property
        def yz(self):
            return len2_type(*self._values[1:3])

        @yz.setter
        def yz(self, value):
            if not isinstance(value, len2_type):
                raise TypeError(value)
            self._values[1:3] = value._values[:]

        class_dict["yz"] = yz

    if count == 3:

        @property
        def xyz(self):
            return self.__class__(*self._values[0:3])

        @xyz.setter
        def xyz(self, value):
            if not isinstance(value, self.__class__):
                raise TypeError(value)
            self._values[0:3] = value._values[:]

        class_dict["xyz"] = xyz

    if count == 4:

        @property
        def xyz(self):
            return len3_type(*self._values[0:3])

        @xyz.setter
        def xyz(self, value):
            if not isinstance(value, len3_type):
                raise TypeError(value)
            self._values[0:3] = value._values[:]

        class_dict["xyz"] = xyz

        @property
        def w(self):
            return self._values[3]

        @w.setter
        def w(self, value):
            self._values[3] = cast_type(value)

        class_dict["w"] = w

        @property
        def zw(self):
            return len2_type(*self._values[2:4])

        @zw.setter
        def zw(self, value):
            if not isinstance(value, len2_type):
                raise TypeError(value)
            self._values[2:4] = value._values[:]

        class_dict["zw"] = zw

        @property
        def yzw(self):
            return len3_type(*self._values[1:4])

        @yzw.setter
        def yzw(self, value):
            if not isinstance(value, len3_type):
                raise TypeError(value)
            self._values[1:4] = value._values[:]

        class_dict["yzw"] = yzw

        @property
        def xyzw(self):
            return self.__class__(*self._values)

        @xyzw.setter
        def xyzw(self, value):
            if not isinstance(value, self.__class__):
                raise TypeError(value)
            self._values[:] = value._values[:]

        class_dict["xyzw"] = xyzw

    if typestr is not None:
        class_dict["__typestr__"] = typestr

    result = type(name, (object,), class_dict)
    objc._registerVectorType(result)
    return result


vector_float2 = make_type("vector_float2", 0.0, float, 2, typestr=b"<2f>")
vector_float3 = make_type(
    "vector_float3", 0.0, float, 3, vector_float2, typestr=b"<3f>"
)
vector_float4 = make_type(
    "vector_float4", 0.0, float, 4, vector_float2, vector_float3, typestr=b"<4f>"
)

vector_double2 = make_type("vector_double2", 0.0, float, 2, typestr=b"<2d>")
vector_double3 = make_type(
    "vector_double3", 0.0, float, 3, vector_double2, typestr=b"<3d>"
)
vector_double4 = make_type(
    "vector_double4", 0.0, float, 4, vector_double2, vector_double3, typestr=b"<4d>"
)


def limit_short(v):
    v = v & 0xFFFF
    if v & 0x8000:
        v = ~v + 1 & 0xFFFF
        return -v
    else:
        return v


def limit_int(v):
    v = v & 0xFFFFFFFF
    if v & 0x80000000:
        v = ~v + 1 & 0xFFFFFFFF
        return -v
    else:
        return v


def limit_uchar(v):
    return v % 0xFF


def limit_ushort(v):
    return v % 0xFFFF


def limit_uint(v):
    return v % 0xFFFFFFFF


vector_short2 = make_type(
    "vector_short2", 0, operator.index, 2, limitrange=limit_short, typestr=b"<2s>"
)
vector_ushort2 = make_type(
    "vector_ushort2",
    0,
    operator.index,
    2,
    is_signed=False,
    limitrange=limit_ushort,
    typestr=b"<2S>",
)
vector_ushort3 = make_type(
    "vector_ushort3",
    0,
    operator.index,
    3,
    vector_ushort2,
    is_signed=False,
    limitrange=limit_ushort,
    typestr=b"<3S>",
)
vector_ushort4 = make_type(
    "vector_ushort4",
    0,
    operator.index,
    4,
    vector_ushort2,
    vector_ushort3,
    is_signed=False,
    limitrange=limit_ushort,
    typestr=b"<4S>",
)

vector_int2 = make_type(
    "vector_int2", 0, operator.index, 2, limitrange=limit_int, typestr=b"<2i>"
)
vector_int3 = make_type(
    "vector_int3",
    0,
    operator.index,
    3,
    vector_int2,
    limitrange=limit_int,
    typestr=b"<3i>",
)
vector_int4 = make_type(
    "vector_int4",
    0,
    operator.index,
    4,
    vector_int2,
    vector_int3,
    limitrange=limit_int,
    typestr=b"<4i>",
)
vector_uint2 = make_type(
    "vector_uint2",
    0,
    operator.index,
    2,
    is_signed=False,
    limitrange=limit_uint,
    typestr=b"<2I>",
)
vector_uint3 = make_type(
    "vector_uint3",
    0,
    operator.index,
    3,
    vector_uint2,
    is_signed=False,
    limitrange=limit_uint,
    typestr=b"<3I>",
)

vector_uchar16 = make_type(
    "vector_uchar16",
    0,
    operator.index,
    16,
    is_signed=False,
    limitrange=limit_uchar,
    typestr=b"<16C>",
)

simd_int2 = vector_int2
simd_int3 = vector_int3
simd_int4 = vector_int4
simd_uint2 = vector_uint2
simd_uint3 = vector_uint3
simd_float2 = vector_float2
simd_float3 = vector_float3
simd_float4 = vector_float4
simd_double2 = vector_double2
simd_double3 = vector_double3
simd_double4 = vector_double4
simd_short2 = vector_short2
simd_ushort2 = vector_ushort2
simd_ushort3 = vector_ushort3
simd_ushort4 = vector_ushort4
simd_uchar16 = vector_uchar16


matrix_float2x2 = objc.createStructType(
    "matrix_float2x2", b"{_matrix_float2x2=[2<2f>]}", ["columns"]
)
matrix_float3x3 = objc.createStructType(
    "matrix_float3x3", b"{_matrix_float3x3=[3<3f>]}", ["columns"]
)
matrix_float4x3 = objc.createStructType(
    "matrix_float4x3", b"{_matrix_float4x3=[4<3f>]}", ["columns"]
)
matrix_float4x4 = objc.createStructType(
    "matrix_float4x4", b"{_matrix_float4x4=[4<4f>]}", ["columns"]
)
matrix_double4x4 = objc.createStructType(
    "matrix_double4x4", b"{_matrix_double4x4=[4<4d>]}", ["columns"]
)
simd_quatf = objc.createStructType("simd_quatf", b"{_simd_quatf=<4f>}", ["vector"])
simd_quatd = objc.createStructType("simd_quatd", b"{_simd_quatd=<4d>}", ["vector"])

simd_float4x4 = objc.createStructType(
    "simd_float4x4", b"{_simd_float4x4=[4<4f>]}", ["columns"]
)



"""
Support for NSDecimalNumber.

The actual class is defined in Foundation, but having the wrapper
here is much more convenient.
"""
__all__ = ()
import operator

from objc._convenience import addConvenienceForClass
from objc._objc import NSDecimal, lookUpClass

NSDecimalNumber = lookUpClass("NSDecimalNumber")


def decimal_new(cls, value=None):
    if value is None:
        return cls.numberWithInt_(0)

    else:
        if isinstance(value, NSDecimal):
            return cls.decimalNumberWithDecimal_(value)
        elif isinstance(value, NSDecimalNumber):
            return cls.decimalNumberWithDecimal_(value.decimalValue())
        elif isinstance(value, float):
            return cls.numberWithDouble_(value)
        elif isinstance(value, str):
            value = NSDecimal(value)
            return cls.decimalNumberWithDecimal_(value)
        else:
            # The value is either an integer, or
            # invalid (and numberWithLongLong_ will raise
            # TypeError)
            try:
                return cls.numberWithLongLong_(value)

            except ValueError:
                raise TypeError("Value is not a number")


addConvenienceForClass(
    "NSDecimalNumber",
    (
        ("__new__", staticmethod(decimal_new)),
        (
            "__add__",
            lambda self, other: NSDecimalNumber(operator.add(NSDecimal(self), other)),
        ),
        (
            "__radd__",
            lambda self, other: NSDecimalNumber(operator.add(other, NSDecimal(self))),
        ),
        (
            "__sub__",
            lambda self, other: NSDecimalNumber(operator.sub(NSDecimal(self), other)),
        ),
        (
            "__rsub__",
            lambda self, other: NSDecimalNumber(operator.sub(other, NSDecimal(self))),
        ),
        (
            "__mul__",
            lambda self, other: NSDecimalNumber(operator.mul(NSDecimal(self), other)),
        ),
        (
            "__rmul__",
            lambda self, other: NSDecimalNumber(operator.mul(other, NSDecimal(self))),
        ),
        (
            "__truediv__",
            lambda self, other: NSDecimalNumber(
                operator.truediv(NSDecimal(self), other)
            ),
        ),
        (
            "__rtruediv__",
            lambda self, other: NSDecimalNumber(
                operator.truediv(other, NSDecimal(self))
            ),
        ),
        (
            "__floordiv__",
            lambda self, other: NSDecimalNumber(
                operator.floordiv(NSDecimal(self), other)
            ),
        ),
        (
            "__rfloordiv__",
            lambda self, other: NSDecimalNumber(
                operator.floordiv(other, NSDecimal(self))
            ),
        ),
        (
            "__mod__",
            lambda self, other: NSDecimalNumber(operator.mod(NSDecimal(self), other)),
        ),
        (
            "__rmod__",
            lambda self, other: NSDecimalNumber(operator.mod(other, NSDecimal(self))),
        ),
        ("__neg__", lambda self: NSDecimalNumber(operator.neg(NSDecimal(self)))),
        ("__pos__", lambda self: NSDecimalNumber(operator.pos(NSDecimal(self)))),
        ("__abs__", lambda self: NSDecimalNumber(abs(NSDecimal(self)))),
        ("__lt__", lambda self, other: (NSDecimal(self) < other)),
        ("__gt__", lambda self, other: (NSDecimal(self) > other)),
        ("__le__", lambda self, other: (NSDecimal(self) <= other)),
        ("__ge__", lambda self, other: (NSDecimal(self) >= other)),
        ("__eq__", lambda self, other: (NSDecimal(self) == other)),
        ("__ne__", lambda self, other: (NSDecimal(self) != other)),
    ),
)

addConvenienceForClass(
    "NSDecimalNumber",
    (("__round__", lambda self, n=0: NSDecimalNumber(round(NSDecimal(self), n))),),
)
